/*
	File name:	2dto2d.c
	Language:	C
	Type:		Subroutine
	Author:		Gang Liu
	Date:		9/12/98
	Function:	Contains the subroutine 2dto2d_perspective_projection()
			which places a 2D digital image in the camera frame
			at certain location and with certain orientation,
			and project the image to the imaging plane with
			perspective projection.
*/

#include "must.h"

#include "2dto2d_misc.h"

#define	OUT_VALUE	255

static float	bi_linear_interpolation(uchar **data, float y, float x);

int	_2dto2d_perspective_projection(uchar **input, int inrows, int incols,
		float c0, float r0,
		float omega, float phi, float kappa,
		float x0, float y0, float z0,
		float f, float delta_u, float delta_v,
		uchar **(*output), int *outrows, int *outcols)
/*
 Arguments:
	input	--->	the input digital image, used as a 2D array of
			'inrows' rows and 'incols' columns;
	inrows	--->	number of rows of the input image;
	incols	--->	number of columns of the input image;
	c0	--->	the column coordinate of the origin of the object frame;
	r0	--->	the row coordinate of the origin of the object frame;
	omega	--->	the rotation angle in radians about the x axis in
			specifying the object frame relative to the camera
			frame;
	phi	--->	similar to 'omega', but for the y axis;
	kappa	--->	similar to 'omega', but for the z axis;
	x0	--->	the x coordinate of the origin of the object frame
			in the camera frame;
	y0	--->	similar to x0, but for the y coordinate;
	z0	--->	similar to x0, but for the z coordinate;
	f	--->	the location of the imaging plane on the z axis;
	delta_u	--->	pixel size along the u axis on the imaging plane;
	delta_v	--->	similar to u, but for the v axis;
	output	<---	output digital image of the desired projection;
	outrows	<---	number of rows of the output image;
	outcols	<---	number of columns of the output image;

 Actions:
	An object frame is used to specify the pose (location and
	orientation) of the input image in the camera frame, from which 
	the perspective projection onto the imaging plane is to be obtained.

	The image is attached to the a-b plane of the object frame, with the
	column-increasing direction aligned with the a axis of the frame, and
	the row-increasing direction aligned with the b axis. The image is
	placed on the a-b plane in such direction, and with the origin of the
	a-b plane placed at the pixel location (row,column)=(r0,c0).

	The orientation of the object frame relative to the camera frame is
	specified by first rotating the camera frame about its x axis for
	the amount of 'omega', the rotating the rotated frame about its y
	axis for 'phi', and finally rotating the twice rotated frame about
	its z axis for 'kappa'. Then the origin of the last frame is moved
	to (x0,y0,z0) of the camera frame, and resulting in the final object
	frame. Since the input digital image is fully specified relative to
	the object frame, so is true for the camera frame.

	The focal length, or the location of the imaging plane on the z axis
	in the camera frame, is given as 'f', and the digitization on the
	imaging plane is determined by the pixel size specified by 'delta_u'
	and 'delta_v', for the u and v axes, respectively. Therefore, pixel
	shape is rectangular when these two values differ, and square when
	they are equal.

	The four corners of the input image is first projected to the
	imaging plane. The bounding box on the imaging plane of these four
	projected points is obtained, which will be sampled with the
	specified pixel size.
	
	Each sample location within the box is back projected to the object
	frame. If the back projected point falls within the domain of the
	input digital image, bi-linear interpolation will be used to determine
	a gray value, and that value will be used for the said sample locaton
	on the imaging plane. For points whose back projected points fall
	outside of the domain of the input image, a default value (given by
	the macro OUT_VALUE) will be used.

	Then we are done.

	Note: The behavior of the subroutine is unpredicted when the input
	image is placed in such a way that it intersects with the imaging
	plane.

 Return values:
	0	Success
	-1	Memory allocation failure for the output image buffer
*/
{
	float	r[3][3];

	float	a, b, u, v;
	float	corners_a[4], corners_b[4];
	float	corners_u[4], corners_v[4];
	float	umin, umax, vmin, vmax;

	int	i, j, k;
	float	finrows, fincols;
	float	val;

	get_rot_mat(omega, phi, kappa, r);

	corners_a[0]=0-c0;		corners_b[0]=0-r0;
	corners_a[1]=incols-1-c0;	corners_b[1]=0-r0;
	corners_a[2]=incols-1-c0;	corners_b[2]=inrows-1-r0;
	corners_a[3]=0-c0;		corners_b[3]=inrows-1-r0;

	for(k=0;k<4;k++)
		abc2uv(r, x0, y0, z0, f, corners_a[k], corners_b[k], 0.0,
			&(corners_u[k]), &(corners_v[k]));

	umin=umax=corners_u[0]; vmin=vmax=corners_v[0];

	for(k=1;k<4;k++) {
		u=corners_u[k]; v=corners_v[k];
		if(umin>u) umin=u;
		if(umax<u) umax=u;
		if(vmin>v) vmin=v;
		if(vmax<v) vmax=v;
	}

	*outrows=(vmax-vmin)/delta_v + 1.0;
	*outcols=(umax-umin)/delta_u + 1.0;

	*output=(uchar**)alloc2d(*outcols, *outrows, sizeof(uchar));
	if(*output==NULL) return -1;

	finrows=inrows-1; fincols=incols-1;
	for(i=0;i<*outrows;i++)
		for(j=0;j<*outcols;j++) {
			v=vmin+i*delta_v;
			u=umin+j*delta_u;
			uv2ab(r, x0, y0, z0, f, u, v, &a, &b);
			a+=c0; b+=r0;
			if(a<0.0 || a>=fincols || b<0.0 || b>=finrows)
				val=OUT_VALUE;
			else
				val=bi_linear_interpolation(input, b, a);

			(*output)[i][j]=val+0.5;

		}

	return 0;
}

static float	bi_linear_interpolation(uchar **data, float y, float x)
{
	float	a, b, val;
	int	i, j;

	i=y;
	j=x;
	b=y-i;
	a=x-j;
	val=	  (1.0-b) * ( (1.0-a)*data[i  ][j] + a*data[i  ][j+1])
		+       b * ( (1.0-a)*data[i+1][j] + a*data[i+1][j+1]);

	return val;
}

#undef	OUT_VALUE
