/*
	File name:	mosaicpgm.c
	Language:	C
	Author:		Gang Liu
	Date:		07/27/97
	Type:		main program
	Function:	Given a series sub-images, each with its upper-left
			corner specified relative to some origin, this utility
			produces an output image with the sub-images placed
			the way they are specified.
			
			The pixel values of the area occupied by the
			sub-images take those of the sub-images; unoccupied
			areas will take value 0. If the sub-images overlap,
			the later sub-images will overwrite former ones.

			The size of the output image will be that of the
			bounding box of all the sub-images.

			If you want to have an underlying background for the
			sub-images, first create an image of desired size and
			desired values, then use it as the first sub-image
			with its upper-left corner specified to (0,0).

*/
#include <stdio.h>
#include <malloc.h>
#include <stdlib.h>

#include "must.h"
#include "pgm.h"

void	usage(char *cmd);

int	main(int argc, char *argv[])
{
	int	i, j, k;

	int	inputs;
	char	*tmpname, *outname;

	uchar	**tmpimage;
	int	tmpwidth, tmpheight;

	uchar	**outimage;
	int	width, height;

	int	x, y, *xlocs, *ylocs, left, top, right, bottom;


	if(argc < 5) {
		usage(argv[0]);
		return -1;
	}

	outname = argv[argc-1];
	inputs = (argc-2) / 3;	/* less command name, input and output names */

	printf("%d input sub-images to produce output %s\n", inputs, outname);

	xlocs=(int*)malloc(inputs*sizeof(int));
	ylocs=(int*)malloc(inputs*sizeof(int));
	if(xlocs==NULL || ylocs==NULL) myerror(-2, stderr, "No memory.\n");

	left=right=top=bottom=0;
	for(k=0;k<inputs;k++) {
		/* name and location of the k-th input sub-image */
		tmpname  = argv[1+k*3];
		xlocs[k] = x = atoi( argv[1+k*3+1] );
		ylocs[k] = y = atoi( argv[1+k*3+2] );
		i=ReadPGM(tmpname, &tmpwidth, &tmpheight, &tmpimage);
		if(i) {
			fprintf(stderr, "For %s:\n", tmpname);
			myerror(-3, stderr, "Error reading input image.\n");
		}
		free2d((void**)tmpimage);

		/* calculate the size of the output image */
		if(left>x) left=x;
		if(right<x+tmpwidth-1) right=x+tmpwidth-1;
		if(top>y) top=y;
		if(bottom<y+tmpheight-1) bottom=y+tmpheight-1;
	}

	width=right-left+1;
	height=bottom-top+1;

	outimage=(uchar**)alloc2d(width, height, sizeof(uchar));
	if(outimage==NULL) myerror(-4, stderr, "No memory.\n");

	for(i=0;i<height;i++)
		for(j=0;j<width;j++) outimage[i][j]=0;

	for(k=0;k<inputs;k++) {
		tmpname  = argv[1+k*3];
		i=ReadPGM(tmpname, &tmpwidth, &tmpheight, &tmpimage);
		if(i) {
			fprintf(stderr, "For %s:\n", tmpname);
			myerror(-5, stderr, "Error reading input image.\n");
		}

		x=xlocs[k];
		y=ylocs[k];
		for(i=0;i<tmpheight;i++)
			for(j=0;j<tmpwidth;j++)
				outimage[i+y-top][j+x-left]=tmpimage[i][j];

		free2d((void**)tmpimage);
	}

	i=WritePGM(outname, width, height, outimage);
	if(i) {
		fprintf(stderr, "For output %s:\n", outname);
		myerror(-6, stderr, "Error writing output image.\n");
	}

	free2d((void**)outimage);
	free(xlocs);
	free(ylocs);

	return 0;
}

void	usage(char *cmd)
{
fprintf(stderr, "\nFunction of %s:\n\n", cmd);
fprintf(stderr,
"Given a series sub-images, each with its upper-left corner specified\n\
relative to some origin, this utility produces an output image with the\n\
sub-images placed the way they are specified.\n\
\n\
The pixel values of the area occupied by the sub-images take those of\n\
the sub-images; unoccupied areas will take value 0. If the sub-images\n\
overlap, the later sub-images will overwrite former ones.\n\
\n\
The size of the output image will be that of the bounding box of all the\n\
sub-images.\n\
\n\
If you want to have an underlying background for the sub-images, first\n\
create an image of desired size and desired values, then use it as the\n\
first sub-image with its upper-left corner specified to (0,0).\n\n");

fprintf(stderr, "Usage: %s <fi1 x1 y1> [<fi2 x2 y2> ...] <fo>\n", cmd);
fprintf(stderr, "\t fi1 - name of first input sub-image file (PGM);\n");
fprintf(stderr,
"\t x1  - column position of the upper-left corner of the 1st sub-image;\n");
fprintf(stderr,
"\t y1  - row    position of the upper-left corner of the 1st sub-image;\n");
fprintf(stderr, "\t fo  - name of the output image file (PGM);\n");
fprintf(stderr, "\n");
fprintf(stderr, "\t <fi2 x2 y2> ... \t additional sub-images;\n");
fprintf(stderr, "\n");
return;
}
