/*
	File name:	pseudocolor.c
	Type:		C main file.
	Author:		Gang Liu
	Date:		01/08/98
	Function:	Create a RGB image in PPM format from one PGM image
			by applying pseudo-color to it.
*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "must.h"
#include "pgm.h"
#include "ppm.h"

#define	LEVELS	256
static	float	lut[LEVELS][3];

void	usage(FILE *fp, char *cmd);
static	int	get_lut_jet(float lut[][3], int levels);
static	int	get_lut_modified_jet(float lut[][3], int levels);
static	int	get_lut_cyclic(float lut[][3], int levels);
static	int	get_lut_heatscale(float lut[][3], int levels);

int	main(int argc, char *argv[])
{
	int	width, height;
	uchar	**image;

	int	choice;
	char	*inname, *outname;
	FILE	*fp;

	uchar	c, c1r, c1g, c1b, c2r, c2g, c2b, err;

	int	i, j;
	int	(*get_lut)(float lut[][3], int levels);

	if(argc<4) {
		usage(stderr, argv[0]);
		return -1;
	}

	i=1;
	choice=atoi(argv[i++]);
	inname=argv[i++];
	outname=argv[i++];

	switch(choice) {
		case 1:
			get_lut=get_lut_jet;
			break;
		case 2:
			get_lut=get_lut_modified_jet;
			break;
		case 3:
			get_lut=get_lut_cyclic;
			break;
		case 4:
			get_lut=get_lut_heatscale;
			break;
		default:
			usage(stderr, argv[0]);
			return -1;
	}


	i=get_lut(lut, LEVELS);
	if(i) myerror(i, stderr, "get_lut() failed.");

	/*
	for(i=0;i<LEVELS;i++)
		printf("%e %e %e\n", lut[i][0], lut[i][1], lut[i][2]);
	*/

	i=ReadPGM(inname, &width, &height, &image);
	if(i) myerror(-2, stderr, "Error reading input image.\n");

	fp=fopen(outname, "w");
	if(fp==NULL) myerror(-3, stderr, "Error openning output file.\n");

	fprintf(fp, "P6\n");
	fprintf(fp, "%d %d\n", width, height);
	fprintf(fp, "255\n");

	err=0;
	for(i=0;i<height;i++)
		for(j=0;j<width;j++) {
			c=image[i][j];
			c1r=lut[c][PPM_RED];
			c1g=lut[c][PPM_GREEN];
			c1b=lut[c][PPM_BLUE];
			c2r=fputc(c1r, fp);
			c2g=fputc(c1g, fp);
			c2b=fputc(c1b, fp);
			if(c1r!=c2r || c1g!=c2g || c1b!=c2b) err=1;
		}
	err |= fclose(fp);
	if(err) myerror(-4, stderr, "Error writing output data.\n");

	free2d((void**)image);

	return 0;
}

void	usage(FILE *fp, char *cmd)
{
fprintf(fp, "Usage:\n\t %s <LUT> <input PGM> <output PPM>\n", cmd);
fprintf(fp,
"where\n\
\t <LUT> is the choice for which color look-up table to use; choices are\n\
\t\t 1 -- \"jet\" table used by Matlab;\n\
\t\t 2 -- \"jet\" with first 8 entries replaced by 8 pure colors;\n\
\t\t 3 -- \"cyclic\" table, with repeating patterns of 8 pure colors;\n\
\t\t 4 -- \"heatscale\" table stolen from http://cis.jhu.edu;\n\
\t <input PGM> is the name of the input PGM image file;\n\
\t <output PGM> is the name of the output PGM image file.\n\
\n");
}

static	int	get_lut_jet(float lut[][3], int levels)
{
	int	i, j;
	int	len1, len2, len3, len4, len5;

/* this color map is created by immitating the Matlab's 'jet' colormap */
	if(levels!=256) return -1;

	len1=32; len2=64; len3=64; len4=64; len5=32;
	j=0;

	for(i=0;i<len1;i++) {
		lut[j+i][PPM_RED]=0.0;
		lut[j+i][PPM_GREEN]=0.0;
		lut[j+i][PPM_BLUE]=i/(float)len1;
	}
	j+=len1;

	for(i=0;i<len2;i++) {
		lut[j+i][PPM_RED]=0.0;
		lut[j+i][PPM_GREEN]=i/(float)len2;
		lut[j+i][PPM_BLUE]=(len1-1)/(float)len1;
	}
	j+=len2;

	for(i=0;i<len3;i++) {
		lut[j+i][PPM_RED]=i/(float)len3;
		lut[j+i][PPM_GREEN]=(len2-1)/(float)len2;
		lut[j+i][PPM_BLUE]=(len3-1-i)/(float)len3;
	}
	j+=len3;

	for(i=0;i<len4;i++) {
		lut[j+i][PPM_RED]=(len3-1)/(float)len3;
		lut[j+i][PPM_GREEN]=(len4-1-i)/(float)len4;
		lut[j+i][PPM_BLUE]=0.0;
	}
	j+=len4;

	for(i=0;i<len5;i++) {
		lut[j+i][PPM_RED]=(len5-1-i/2.0)/(float)(len5);
		/*
		lut[j+i][PPM_GREEN]=0.0;
		lut[j+i][PPM_BLUE]=0.0;
		*/
		lut[j+i][PPM_GREEN]=i/(float)len4;
		lut[j+i][PPM_BLUE]=0.0;
	}
	j+=len5;

	for(i=0;i<levels;i++)
		for(j=0;j<3;j++) lut[i][j]*=levels;

	return 0;
}

static	int	get_lut_1(float lut[][3], int levels)
{
	int	i, j;
	int	len1, len2, len3, len4, len5;

/* this color map is created by immitating the Matlab's 'jet' colormap */
	if(levels!=256) return -1;

	len1=32; len2=64; len3=64; len4=64; len5=32;
	j=0;

	for(i=0;i<len1;i++) {
		lut[j+i][PPM_RED]=0.0;
		lut[j+i][PPM_GREEN]=0.0;
		lut[j+i][PPM_BLUE]=i/(float)len1;
	}
	j+=len1;

	for(i=0;i<len2;i++) {
		lut[j+i][PPM_RED]=0.0;
		lut[j+i][PPM_GREEN]=i/(float)len2;
		lut[j+i][PPM_BLUE]=(len1-1)/(float)len1;
	}
	j+=len2;

	for(i=0;i<len3;i++) {
		lut[j+i][PPM_RED]=i/(float)len3;
		lut[j+i][PPM_GREEN]=(len2-1)/(float)len2;
		lut[j+i][PPM_BLUE]=(len3-1-i)/(float)len3;
	}
	j+=len3;

	for(i=0;i<len4;i++) {
		lut[j+i][PPM_RED]=(len3-1)/(float)len3;
		lut[j+i][PPM_GREEN]=(len4-1-i)/(float)len4;
		lut[j+i][PPM_BLUE]=i/2.0/(float)len3;
	}
	j+=len4;

	for(i=0;i<len5;i++) {
		lut[j+i][PPM_RED]=(len5-1-i/2.0)/(float)(len5);
		lut[j+i][PPM_GREEN]=i/(float)len5;
		lut[j+i][PPM_BLUE]=(len3-1)/2.0/(float)len3;
	}
	j+=len5;

	for(i=0;i<levels;i++)
		for(j=0;j<3;j++) lut[i][j]*=levels;

	for(j=0;j<3;j++) {
		lut[0][j]=0;
		lut[levels-1][j]=levels-1;
	}

	return 0;
}

static	int	get_lut_modified_jet(float lut[][3], int levels)
{
	int	i, j;
	int	len;

/* this color map is created by immitating the Matlab's 'jet' colormap */

/* with the first 8 entries modified for some "pure colors" */

	if(levels!=256) return -1;

	len=levels/4;

	j=0;
	for(i=0;i<len;i++) {
		lut[j+i][PPM_RED]=0.0;
		lut[j+i][PPM_GREEN]=i/(float)len;
		lut[j+i][PPM_BLUE]=(len-1)/(float)len;
	}
	j+=len;

	for(i=0;i<len;i++) {
		lut[j+i][PPM_RED]=0.0;
		lut[j+i][PPM_GREEN]=(len-1)/(float)len;
		lut[j+i][PPM_BLUE]=(len-1-i)/(float)len;
	}
	j+=len;

	for(i=0;i<len;i++) {
		lut[j+i][PPM_RED]=i/(float)len;
		lut[j+i][PPM_GREEN]=(len-1)/(float)len;
		lut[j+i][PPM_BLUE]=0.0;
	}
	j+=len;

	for(i=0;i<len;i++) {
		lut[j+i][PPM_RED]=(len-1)/(float)len;
		lut[j+i][PPM_GREEN]=(len-1-i)/(float)len;
		lut[j+i][PPM_BLUE]=0.0;
	}
	j+=len;

	for(i=0;i<levels;i++)
		for(j=0;j<3;j++) lut[i][j]*=levels;

	for(j=0;j<3;j++) {
		lut[0][j]=0;
		lut[levels-1][j]=levels-1;
	}

	/* get black, red, green, blue, cyan, magenta, yellow, white */
	for(i=0;i<8;i++)
		for(j=0;j<3;j++) lut[i][j]=0;

	i=1;
	lut[i][PPM_RED]=levels-1; i++;
	lut[i][PPM_GREEN]=levels-1; i++;
	lut[i][PPM_BLUE]=levels-1; i++;
	lut[i][PPM_GREEN]=lut[i][PPM_BLUE]=levels-1; i++;
	lut[i][PPM_RED]=lut[i][PPM_BLUE]=levels-1; i++;
	lut[i][PPM_RED]=lut[i][PPM_GREEN]=levels-1; i++;
	lut[i][PPM_RED]=lut[i][PPM_GREEN]=lut[i][PPM_BLUE]=levels-1; i++;

	return 0;
}

static	int	get_lut_cyclic(float lut[][3], int levels)
{
	int	i, j;

/* the LUT with the repeating 8 "pure colors" */

	if(levels!=256) return -1;

	/* get black, red, green, blue, cyan, magenta, yellow, white */
	for(i=0;i<levels;i++)
		for(j=0;j<3;j++) lut[i][j]=0;

	for(i=0;i<levels;) {
		i++;
		lut[i][PPM_RED]=levels-1; i++;
		lut[i][PPM_GREEN]=levels-1; i++;
		lut[i][PPM_BLUE]=levels-1; i++;
		lut[i][PPM_GREEN]=lut[i][PPM_BLUE]=levels-1; i++;
		lut[i][PPM_RED]=lut[i][PPM_BLUE]=levels-1; i++;
		lut[i][PPM_RED]=lut[i][PPM_GREEN]=levels-1; i++;
		lut[i][PPM_RED]=lut[i][PPM_GREEN]=lut[i][PPM_BLUE]=levels-1;i++;
	}

	return 0;
}

static	int	get_lut_heatscale(float lut[][3], int levels)
{
	int	i;

/* the LUT for the "heatscale" from the CIS of JHU (http://cis.jhu.edu) */

	if(levels!=256) return -1;

	for(i=0;i<64;i++) {
		lut[i][PPM_RED]=2*i;
		lut[i][PPM_GREEN]=0;
		lut[i][PPM_BLUE]=0;
	}

	for(i=64;i<128;i++) {
		lut[i][PPM_RED]=2*i;
		lut[i][PPM_GREEN]=2*(i-64);
		lut[i][PPM_BLUE]=0;
	}

	for(i=128;i<192;i++) {
		lut[i][PPM_RED]=254;
		lut[i][PPM_GREEN]=2*(i-64);
		lut[i][PPM_BLUE]=2*(i-128);
	}

	for(i=192;i<256;i++) {
		lut[i][PPM_RED]=254;
		lut[i][PPM_GREEN]=254;
		lut[i][PPM_BLUE]=2*(i-128);
	}

	return 0;
}

