/*
	File name:	rmtinycomppgm.c
	Type:		C main file.
	Author:		Gang Liu
	Date:		12/21/97
	Function:	Given an input binary file in PGM format, find its
			foreground pixels which belong to components of sizes
			smaller than the given threshold; assign the label 0,
			the label for the background to these pixels; and give
			the remaining foreground pixels a connected component
			labeling. The result is given in an output image file
			also in PGM format.
			
			Note:
			  1.  0 in the input image stands for the background,
				and all none-zero values treated as foreground;
			  2.  a macro WRAP is defined;
				if there are more than WRAP connected components
				in the input image, the symbols are wrapped
				back into the range between 1 and WRAP, with 0
				reserved for background.
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#define	WRAP	255

#include "must.h"
#include "pgm.h"
#include "connect.h"

void	usage(FILE *fp, char *cmd);

int	main(int argc, char *argv[])
{
	char	verbose;
	char	*inname, *outname;
	int	thres;

	int	width, height;
	uchar	**image;

	int	neighbors;
	int	num_components;
	int	**labels, tmp;

	int	i, j;

	if(argc<5) {
		usage(stderr, argv[0]);
		return -1;
	}

	neighbors=atoi(argv[1]);
	inname=argv[2];
	thres=atoi(argv[3]);
	outname=argv[4];

	verbose=0;
	if(argc>=6 && toupper(argv[5][0])=='V') verbose=1;

	if(neighbors!=4 && neighbors!=8) neighbors=8;
	if(thres<0) thres=0;

	i=ReadPGM(inname, &width, &height, &image);
	if(i) myerror(-2, stderr, "Error reading input image.\n");

	labels=(int**)alloc2d(width, height, sizeof(int));
	if(labels==NULL) myerror(-3, stderr, "Not enough memory.\n");

	i=connected_component(image[0], width, height, neighbors,
				labels[0], &num_components);
	if(i) myerror(i, stderr, "connected_component() failed.\n");

	i=remove_component(labels[0], width, height, &num_components, thres);
	if(i) myerror(i, stderr, "remove_component() failed.\n");

	for(i=0;i<height;i++)
		for(j=0;j<width;j++) {
			tmp=labels[i][j];
			if(tmp>WRAP) {
				tmp %= WRAP;
				if(!tmp) tmp=WRAP; /* multiples of WRAP */
			}
			image[i][j]=tmp;
		}

	i=WritePGM(outname, width, height, image);
	if(i) fprintf(stderr, "Error writing output.\n");

	if(verbose)
	printf("Using %d-connectivity and %d connected components found.\n",
		neighbors, num_components);

	if(num_components>WRAP)
		fprintf(stderr,
"Warning: number of connected components exceeded %d; extras wrapped.\n", WRAP);

	free2d((void**)image);
	free2d((void**)labels);

	if(i) i=-2;	/* signaling error in writing output file */
	return i;
}

void	usage(FILE *fp, char *cmd)
{
fprintf(fp, "Usage:\n\t %s <8|4> <PGM1> <thres> <PGM2> [v]\n", cmd);
fprintf(fp, "<8|4>  use 8- or 4-connectivity;\n");
fprintf(fp, "<PGM1> is input PGM file;\n");
fprintf(fp, "<thres> the size of smallest connected component of interest;\n");
fprintf(fp, "<PGM2> is output PGM file;\n");
fprintf(fp, "[v]    verbose.\n");
fprintf(fp, "\n");
return;
}
